#!/usr/bin/env python3

"""Shared stuff"""
# pylint: disable=too-many-arguments

from collections.abc import Mapping, Sequence
from contextlib import suppress
from dataclasses import dataclass, field
from pathlib import Path

import toml
from sf2utils.sf2parse import Sf2File

NestedConfig = Mapping[str, object]


@dataclass
class MidiEvent:
    """An internal application event"""

    status: int
    data1: int
    data2: int
    data3: int
    timestamp: int


@dataclass
class SfPreset:
    """Configuration structure a SoundFont preset"""

    # channel: None | int = None
    name: str
    short: None | str = None
    bank: None | int = None
    preset: None | int = None


@dataclass
class SfConfig:
    """Stores meta info about SoundFont files"""

    filepath: Path
    instruments: Sequence[SfPreset]
    info: None | Mapping[str, str] = None

    def __init__(self, filepath: str | Path, instruments: None | Sequence[NestedConfig] = None):
        self.filepath = Path(filepath)
        if instruments:
            self.instruments = [SfPreset(**element) for element in instruments]
        else:
            with open(filepath, "rb") as sf2_file:
                sf2 = Sf2File(sf2_file)
                self.info = {
                    key.decode(): value.decode().strip("\0") if isinstance(value, bytes) else value
                    for key, value in sf2.raw.info.items()
                }
                self.instruments = [
                    SfPreset(
                        # channel=None,
                        name=preset.name,
                        short=None,
                        bank=preset.bank,
                        preset=preset.preset,
                    )
                    for preset in sorted(
                        (_p for _p in sf2.presets if _p.name != "EOP"),
                        key=lambda p: (p.bank, p.preset),
                    )
                ]

    @staticmethod
    def fromMap(args: Mapping[str, object]) -> "SfConfig":
        """Factory function"""
        if args.keys() != {"filepath", "instruments"}:
            raise RuntimeError(f"Bad SoundFont mapping keys: {set(args.keys())}")
        return SfConfig(args["filepath"], args["instruments"])


@dataclass
class MidiConfig:
    """Configuration structure for a MIDI backend"""

    backend: str = "fluidsynth"
    audio: NestedConfig = field(default_factory=dict)
    reverb: None | object = None
    chorus: None | object = None
    sf_load: None | Sequence[SfConfig] = None

    def __init__(
        self,
        backend: None | str = None,
        sf_load: None | Sequence[NestedConfig] = None,
        audio: None | NestedConfig = None,
        reverb: None | object = None,
        chorus: None | object = None,
    ) -> None:
        if backend:
            self.backend = backend
        if sf_load is not None:
            self.sf_load = [SfConfig.fromMap(element) for element in sf_load]
        self.audio = {**{"period-size": 256}, **(audio or {})}
        if reverb:
            self.reverb = reverb
        if chorus:
            self.chorus = chorus


@dataclass
class PriConfig:
    """Stores pocketrockit configuration"""

    midi: MidiConfig = field(default_factory=MidiConfig)

    def __init__(self, data: Path):
        with suppress(FileNotFoundError):
            with open(data, encoding="utf-8") as toml_file:
                # TomlDecodeError
                raw = toml.load(toml_file)
                self.midi = MidiConfig(**(raw.get("pocketrockit") or {}).get("midi") or {})
                return
        self.midi = MidiConfig()



#!/usr/bin/env python3

"""Brings together all publicly needed symbols"""

from pocketrockit.decorated import echo, midiseq, player, track
from pocketrockit.engine import EMPTY_COMMAND, Env

__all__ = ["Env", "midiseq", "player", "track", "EMPTY_COMMAND", "echo"]



#!/usr/bin/env python3

"""POCKETROCKIT
"""

import importlib.metadata
import logging
import shutil
import sys
from argparse import ArgumentParser
from argparse import Namespace as Args
from collections.abc import Sequence
from contextlib import suppress
from pathlib import Path

from pocketrockit.commons import SfConfig
from pocketrockit.engine import load_module
from pocketrockit.misc import setup_logging


def parse_args(argv: Sequence[str] | None = None) -> Args:
    """Cool git like multi command argument parser"""
    parser = ArgumentParser(__doc__)
    parser.add_argument("--verbose", "-v", action="store_true")

    parser.set_defaults(func=lambda *_: parser.print_usage())
    subparsers = parser.add_subparsers(help="available commands", metavar="CMD")

    parser_info = subparsers.add_parser("help")
    parser_info.set_defaults(func=lambda *_: parser.print_help())

    parser_info = subparsers.add_parser("info")
    parser_info.set_defaults(func=fn_info)

    parser_play = subparsers.add_parser("play", aliases=["p", "run"])
    parser_play.set_defaults(func=fn_play)
    parser_play.add_argument("path", type=Path, default=".", nargs="?")

    parser_init = subparsers.add_parser("init")
    parser_init.set_defaults(func=fn_init)
    parser_init.add_argument("path", type=Path, default=".", nargs="?")

    parser_list_sf2 = subparsers.add_parser("list-sf2")
    parser_list_sf2.set_defaults(func=fn_list_sf2)
    parser_list_sf2.add_argument("path", type=Path, default=".")

    return parser.parse_args(argv)


def logger() -> logging.Logger:
    """Named logger"""
    return logging.getLogger("pr.cli")


def extract_version() -> str:
    """Returns either the version of installed package or the one
    found in nearby pyproject.toml"""
    with suppress(FileNotFoundError, StopIteration):
        with open(
            Path(__file__).parent.parent / "pyproject.toml", encoding="utf-8"
        ) as pyproject_toml:
            version = (
                next(line for line in pyproject_toml if line.startswith("version"))
                .split("=")[1]
                .strip("'\"\n ")
            )
            return f"{version}-dev"
    return importlib.metadata.version(__name__.split(".", maxsplit=1)[0])


__version__ = extract_version()


def shorten_home(path: Path | str) -> Path:
    """Reverse of expanduser"""
    return Path(Path(path).as_posix().replace(str(Path.home()), "~"))


def fn_info(_args: Args) -> None:
    """Entry point `info`"""
    print(f"Version: {__version__} (at {shorten_home(Path(__file__).parent)})")
    print(
        f"Python: {'.'.join(map(str, sys.version_info[:3]))}"
        f" (at {shorten_home(sys.executable)})"
    )
    # Show MIDI info
    # Show Info about SoundFonts
    # Show Info about local track files


def fn_list_sf2(args: Args) -> None:
    """Entry point `list-sf2`"""

    sf2 = SfConfig(args.path)
    print("Use these entries in your .toml file:")
    print("")
    print("[[pocketrockit.midi.sf_load]]")
    print(f'filepath = "{args.path}"')
    print("instruments = [")

    for preset in sf2.presets:
        print(
            f'    {{short="", name="{preset.name}"'
            f", bank={preset.bank}, preset={preset.preset}}},"
        )
    print("]")


def fn_play(args: Args) -> None:
    """Entry point `play`/`run`"""
    setup_logging()
    logger().debug("Args: %s", {k: str(v) for k, v in args.__dict__.items()})
    load_module(args.path)


def fn_init(args: Args) -> None:
    """Entry point `init`"""
    template_target = args.path / "template.py"
    print(
        "\n".join(
            line.lstrip().split("|", 1)[-1]
            for line in """
            Make sure the following requirements are met:
            [ ] FluidSynth is installed and running
            [ ] SoundFont files `instrumental.sf2` and `drums.sf2` are present
            |   (see https://projects.om-office.de/frans/pocketrockit/-/blob/main/Readme.md)
            |   - test with `fluidsynth -i -a pipewire -z 256 -g 0.5 instrumental.sf2 <MIDI_FILE>`
            [ ] A pocketrockit track file is available
            """.split(
                "\n"
            )
        )
    )

    if template_target.exists():
        print(f"{template_target} already exists - won't be overwritten")
    else:
        shutil.copy(Path(__file__).parent / "examples" / "template.py", template_target)
        print(f"A file {template_target} has been made available - take it as a starting point!")


def main() -> int:
    """Entry point for everything else"""
    (args := parse_args()).func(args)
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""POCKETROCKIT engine
"""

import asyncio
import concurrent
import logging
import threading
import time
from asyncio import sleep as async_sleep
from collections.abc import (
    Generator,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sequence,
)
from contextlib import suppress
from dataclasses import dataclass, field
from importlib.machinery import SourceFileLoader
from importlib.util import module_from_spec, spec_from_file_location
from itertools import chain, count
from pathlib import Path
from signal import SIGINT, SIGTERM
from types import ModuleType

# pylint: disable=too-many-locals
# pylint: disable=too-many-instance-attributes
# pylint: disable=fixme
import toml
from asyncinotify import Inotify, Mask

from .commons import MidiEvent, PriConfig
from .midi_io import midi_input_device, midi_output_device
from .misc import Singleton, colored, error, keyboard_reader, setup_logging, watchdog

MidiCmd = tuple[int, int, None | int]
CmdTrace = tuple[int, int, int]
Command = tuple[CmdTrace, str, Sequence[MidiCmd]]
Player = Generator[Command, None | int, None]


EMPTY_COMMAND: Command = (-1, -1, -1), "", []


@dataclass
class Env:
    """Some global environment"""

    # number of ticks per line - a line has typically 4 beats,
    # and should be dividable a number of times without loss of precision,
    # so we combine a number of typical dividers
    step_size: int = 4 * 16 * 3 * 5  # 960
    bpm: int = 120


@dataclass
class Stage(metaclass=Singleton):
    """Access to the currently played music"""

    env: Env = field(default_factory=Env)
    active: bool = False
    # On startup we want the current track definition to be free from errors. Verified will
    # be set once we successfully read from all generators
    verified: bool = False
    players: Mapping[str, Player] = field(default_factory=dict)
    new_track: None | MutableMapping[str, Player] = None
    file_to_track: str = ""
    tick: int = -1
    pause: bool = False
    step: bool = False
    mod: int = 0
    modifiers: MutableSequence[object] = field(default_factory=list)


def load_module(filepath: str | Path) -> ModuleType:
    """(Re)loads a python module specified by @filepath"""
    logger().debug("load module '%s' from '%s'", Path(filepath).stem, filepath)
    spec = spec_from_file_location(Path(filepath).stem, filepath)
    if not (spec and spec.loader):
        raise RuntimeError("Could not load")
    module = module_from_spec(spec)
    assert module
    assert isinstance(spec.loader, SourceFileLoader)
    loader: SourceFileLoader = spec.loader
    # here the actual track definition takes place
    loader.exec_module(module)
    return module


def note_generator() -> Iterator[Iterable[MidiCmd]]:
    """Brings all players together and cleans up played notes"""
    active_notes: MutableMapping[tuple[int, int], int] = {}
    stage = Stage()

    for tick in count():
        stage.tick = tick

        if tick % stage.env.step_size == 0:
            logger().debug(
                "== tick: %d, step: %d, step_size: %d ==",
                tick,
                tick / stage.env.step_size,
                stage.env.step_size,
            )

        all_commands: MutableSequence[MidiCmd] = []
        for name, stream in stage.players.items():
            try:
                source_pos, origin, commands = next(stream)
                if commands:
                    logger().debug("%s: %s:%s => %s", name, source_pos, origin, commands)
                if commands:
                    all_commands.extend([(a, b + stage.mod, c) for a, b, c in commands])
            except StopIteration:
                pass
            except Exception as exc:  # pylint: disable=broad-except
                if not stage.verified:
                    raise
                error(f"exception in play: {exc}")

        stoppers = []
        for (active_channel, active_note), store_tick in list(active_notes.items()):
            # send noteoff for any still played note
            if tick - store_tick > stage.env.step_size // 4:
                stoppers.append((active_channel, active_note, None))
                del active_notes[active_channel, active_note]
        for channel, note, _velocity in all_commands or []:
            # print((channel, note), active)
            if (channel, note) in active_notes:
                # send noteoff for next note to be played
                stoppers.append((channel, note, None))
                del active_notes[channel, note]

            active_notes[channel, note] = tick

        # print(tick, active_notes)
        yield chain(stoppers, all_commands)


@watchdog
async def music_loop() -> None:
    """Main loop collecting instructions to send it to MIDI"""
    stage = Stage()

    config = PriConfig("pocketrockit.toml")

    with midi_output_device(config.midi) as midi_out:
        for i in count():
            # Wait for players to emerge the first time (otherwise `tick` won't start at 0).
            # Inform the user, if there is nothing there yet.
            if stage.players:
                break
            if i % 10 == 0:
                logger().info("no `players` defined yet..")
            await async_sleep(0.1)

        # time progressor - for drift-less looping
        time_ladder = time.time()

        for notes in note_generator():
            for channel, note, velocity in notes:
                try:
                    if velocity is None:
                        # print("OF", (channel, note))
                        midi_out.noteoff(channel, note)
                    else:
                        stage.step = False
                        # print("ON", (channel, note, velocity))
                        midi_out.noteon(channel, note, velocity)
                except Exception as exc:  # pylint: disable=broad-except
                    error(f"Unhandled exception in MIDI playback: {exc}")
            while True:
                # make me absolute please

                # A BPM of 60 typically means we have a duration of 1sec per quater note.
                # A typical 4/4 line would then take 4sec. Env.step_size is the number of ticks
                # per line, so a step_size of 48 would lead to a duration per tick of
                # 4 / 48 sec
                time_ladder += 60 / stage.env.bpm * 4 / stage.env.step_size
                await async_sleep(max(0, time_ladder - time.time()))

                if not stage.pause or stage.step:
                    break


@watchdog
async def watch_changes() -> None:
    """Watches for file changes in track definition file and reload on change"""
    stage = Stage()
    load_module(stage.file_to_track)
    with Inotify() as inotify:
        inotify.add_watch(
            Path(stage.file_to_track).parent,
            mask=Mask.CLOSE_WRITE | Mask.MOVED_TO | Mask.CREATE,
        )
        async for event in inotify:
            # Fixme - currently CLOSE_WRITE might happen twice - looks like we need a
            #         timer mechanism here
            # Fixme - we listen for all files now, regardless whether or not we need it
            #         is that a problem?
            assert event.path
            if not event.path.is_file() or event.path.stem.startswith("."):
                continue
            logger().debug("\033[0;33mpath: %s, mask: %r", event.path, event.mask)
            try:
                load_module(stage.file_to_track)
            except Exception as exc:  # pylint: disable=broad-except
                error(f"Unhandled exception in watch_changes: {exc}")


def logger() -> logging.Logger:
    """Named logger"""
    return logging.getLogger("pr.engine")


@watchdog
async def handle_keyboard(loop: asyncio.BaseEventLoop, terminator: threading.Event) -> None:
    """Handles key press event"""
    ticks = []
    try:
        async for key in keyboard_reader(loop, terminator):
            if key == " ":
                logger().info(colored("SPACE", "yellow"))
                Stage().pause = not Stage().pause
            elif key == ".":
                logger().info(colored("STEP", "yellow"))
                Stage().pause = True
                Stage().step = True
            elif key == "t":
                ticks.append(Stage().tick)
                logger().info(colored(f"{ticks}", "yellow"))
            else:
                logger().info(colored(key, "yellow"))
        logger().info("Keyboard loop stopped - terminate program")
        terminate(terminator)
    except RuntimeError as exc:
        logger().warning("Could not run keyboard handler: %s", exc)


@watchdog
async def handle_midi_input(loop: asyncio.BaseEventLoop, terminator: threading.Event) -> None:
    """Handles key press event"""
    event_queue: asyncio.Queue[MidiEvent] = asyncio.Queue()
    stage = Stage()
    with concurrent.futures.ThreadPoolExecutor() as pool:

        loop.run_in_executor(
            pool,
            midi_input_device,
            "pygame",
            ["OP-1", "Sylphyo", "USB MIDI Interface", "Midi Through"],
            loop,
            event_queue,
            terminator,
        )

        while True:
            event = await event_queue.get()
            if event.status == 176:
                logger().info("MIDI CTRL    %d => %d", event.data1, event.data2)
                if event.data1 == 1:
                    if event.data2 in {1, 127}:
                        stage.mod += -1 if event.data2 == 127 else 1
            elif event.status == 144:
                logger().info("MIDI NOTEON  %d => %d", event.data1, event.data2)
            elif event.status == 128:
                logger().info("MIDI NOTEOFF %d => %d", event.data1, event.data2)
            else:
                logger().warning("MIDI unknown %s", event)


def terminate(terminator: threading.Event) -> None:
    """Sends a signal to async tasks to tell them to stop"""
    try:
        terminator.set()
        for task in asyncio.all_tasks():
            task.cancel()
        asyncio.get_event_loop().stop()
    except Exception as exc:  # pylint: disable=broad-except
        logger().error("terminator got: %r", exc)


def run() -> None:
    """Runs the pocketrockit event loop forever"""
    setup_logging()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    terminator = threading.Event()

    # https://stackoverflow.com/a/71956673
    _tasks = {
        asyncio.ensure_future(handle_midi_input(loop, terminator)),
        asyncio.ensure_future(handle_keyboard(loop, terminator)),
        asyncio.ensure_future(music_loop()),
        asyncio.ensure_future(watch_changes()),
    }

    try:
        for signal_enum in [SIGINT, SIGTERM]:
            loop.add_signal_handler(signal_enum, lambda: terminate(terminator))
        loop.run_forever()
    finally:
        logger().debug("finally - loop.run_forever()")



#!/usr/bin/env python3

"""Stuff needed to communicate with MIDI devices
"""

import asyncio
import logging
import os
import threading
import time
from collections.abc import Iterable, Iterator, MutableMapping, Sequence
from contextlib import contextmanager, suppress
from typing import ContextManager

import fluidsynth

# pylint: disable=wrong-import-position
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "1"
import pygame.midi

from .commons import MidiConfig, MidiEvent, SfConfig


def logger() -> logging.Logger:
    """Named logger"""
    return logging.getLogger("pr.midi_io")


def choose(choices: Iterable[str], *wishlist: str) -> str:
    """Choose from a list of strings until we found someting matching"""
    for wish in wishlist:
        with suppress(StopIteration):
            return next(name for name in choices if wish in name)
    raise KeyError(f"{wishlist}")


# def read_midi(loop, event_queue, terminator):
# logger().debug(">> read_midi")

# try:
# available_input_ports = set(mido.get_input_names())
# logger().info("Available MIDI input ports: \n%s", "  \n".join(available_input_ports))

# chosen_input = choose(
# available_input_ports, "OP-1", "Sylphyo", "USB MIDI Interface", "Midi Through"
# )
# logger().info("Chosen: %r", chosen_input)

# with mido.open_input(chosen_input) as inport:
# while not terminator.is_set():
# if (event := inport.receive()).type == "clock":
# continue
# asyncio.run_coroutine_threadsafe(event_queue.put(event), loop=loop)
# logger().debug("read_midi: got termination signal")

# except Exception as exc:
# logger().error("Unhandled exception in read_midi thread: %s", exc)

# logger().debug("<< read_midi")


@contextmanager
def midi_output_pygame() -> Iterator[pygame.midi.Output]:
    """
    fluidsynth -vv -a pipewire -z 256 -g 2 FluidR3_GM.sf2
    """
    try:
        pygame.midi.init()
        for i in range(pygame.midi.get_count()):
            logger().debug("found MIDI device: %d: %s", i, pygame.midi.get_device_info(i))
        player = pygame.midi.Output(2)
        player.set_instrument(0)
        yield player
    finally:
        del player
        pygame.midi.quit()


class MidiOutputFluidsynth:
    """Wrapps Fluidsynth to create a configurable MIDI player"""

    def __init__(self, config: MidiConfig):
        def raise_on_error(result: int) -> int:
            if result < 0:
                raise RuntimeError("Call to FluidSynth function returned error")
            return result

        self.midi_out = fluidsynth.Synth(gain=2.0, samplerate=44100.0, channels=256)
        for key, value in config.audio.items():
            self.midi_out.setting(f"audio.{key}", value)
        if config.reverb:
            raise_on_error(self.midi_out.set_reverb(**config.reverb))
        if config.chorus:
            raise_on_error(self.midi_out.set_chorus(**config.chorus))

        self.instruments: MutableMapping[int | str, int] = {}

        for sf_load in config.sf_load or [SfConfig("instrumental.sf2")]:
            sf2 = raise_on_error(self.midi_out.sfload(sf_load.filepath.as_posix()))
            for instrument in sf_load.instruments:
                channel = next(i for i in range(256) if i not in self.instruments)
                if instrument.name in self.instruments:
                    raise RuntimeError(f"Name for instrument already taken: {instrument.name}")
                self.instruments[instrument.name] = channel
                if instrument.short:
                    if instrument.short in self.instruments:
                        raise RuntimeError(
                            f"Short name for instrument already taken: {instrument.short}"
                        )
                    self.instruments[instrument.short] = channel
                if channel in self.instruments:
                    raise RuntimeError(f"Channel index for instrument already taken: {channel}")
                self.instruments[channel] = channel
                print(
                    f"ch {channel}: {sf_load.filepath.stem}"
                    f", {instrument.bank}:{instrument.preset} {instrument.name}"
                )
                raise_on_error(
                    self.midi_out.program_select(channel, sf2, instrument.bank, instrument.preset)
                )

        raise_on_error(self.midi_out.start())

    def noteon(self, channel: int, note: int, velocity: int) -> None:
        """noteon wrapper"""
        if channel not in self.instruments:
            logger().error("%s not configured", channel)
        self.midi_out.noteon(self.instruments.get(channel, 1), note, velocity)

    def noteoff(self, channel: int, note: int) -> None:
        """noteoff wrapper"""
        self.midi_out.noteoff(self.instruments.get(channel, 1), note)

    def __enter__(self) -> "MidiOutputFluidsynth":
        return self

    def __exit__(self, *args: object) -> bool:
        self.midi_out.delete()
        return True


def midi_output_device(config: MidiConfig) -> ContextManager[MidiOutputFluidsynth]:
    """Implementation agnostic wrapper for synthesizer context manager"""
    if config.backend == "fluidsynth":
        return MidiOutputFluidsynth(config)
    raise RuntimeError(f"Don't know MIDI backend {config.backend}")


def device_id(choices: Iterable[str]) -> int:
    """Select a pygame midi device id given a list of substring choices"""
    raw_input_devices = tuple(
        (i, pygame.midi.get_device_info(i)) for i in range(pygame.midi.get_count())
    )
    input_devices = {
        i: bname.decode()
        for i, (_, bname, is_input, *_) in raw_input_devices
        if is_input
        if any(c.lower() in bname.decode().lower() for c in choices)
    }
    logger().info("input devices: %s", input_devices)

    for choice in choices:
        for dev_id, dev_name in input_devices.items():
            if choice.lower() in dev_name.lower():
                return dev_id
    raise RuntimeError("Didn't find a device name matching any of the provided conditions")


def midi_input_pygame(
    choices: Iterable[str],
    loop: asyncio.BaseEventLoop,
    event_queue: asyncio.Queue[MidiEvent],
    terminator: threading.Event,
) -> None:
    """Read MIDI input via pygame"""
    if not event_queue:
        return None
    try:
        pygame.midi.init()

        if (input_device_id := device_id(choices)) is None:
            logger().info("No MIDI devices attached")
            return None

        input_device = pygame.midi.Input(input_device_id)
        while not terminator.is_set():
            if not input_device.poll():
                time.sleep(0.01)
                continue
            for status, data1, data2, data3, tick in (
                (_data[0], _data[1], _data[2], _data[3], _tick)
                for _data, _tick in input_device.read(100)
                if isinstance(_data, list)
                if _data != [248, 0, 0, 0]
            ):
                assert isinstance(tick, int)
                asyncio.run_coroutine_threadsafe(
                    event_queue.put(MidiEvent(status, data1, data2, data3, tick)),
                    loop=loop,
                )
    finally:
        pygame.midi.quit()
    return None


def midi_input_device(
    name: str,
    choices: Sequence[str],
    loop: asyncio.BaseEventLoop,
    event_queue: asyncio.Queue[MidiEvent],
    terminator: threading.Event,
) -> None:
    """Implementation agnostic wrapper for MIDI input context manager"""
    try:
        if name:
            return midi_input_pygame(
                choices=choices, loop=loop, event_queue=event_queue, terminator=terminator
            )
    except RuntimeError as exc:
        logger().warning("Could not attach MIDI input device: '%s'", exc)
    return None



#!/usr/bin/env python3
"""You can write some lines here in case you're more into writing than into music"""

# feel free to import any other modules
import random

# you can also write from `pocketrockit import *` - but this is more pythonic
from pocketrockit import Env, midiseq, player, track

# tell formaters not to destroy your ASCII art
# fmt: off


@track
def track_name(env: Env):
    """Put a little poem here, if you want. I'll describe the currently available syntax for
    sequencer pattern used for midiseq():
        - 'C', 'D', 'E'.. 'E3'.. 'fis', 'ges'.. specify notes
        - 'I' .. 'VIII' specify steps relative to a key
        - ' ' (space) separates notes
        - '|' separates 'measures'
        - ,: separate notes in a chord, e.g. 'C,E,F'
        - paranthesis divide rythm
        - +/-: shift half tones
        - ><: shift octaves
    """

    # @env object contains some global stuff - currently: only bpm and step_size
    env.bpm = 100

    key = "A3,A3,A3-"

    # The @player decorator defines one instrumental part played along others
    # Choose any valid function name - functions with same name overwrite each other
    @player
    def metronome():
        # use MIDI numbers directly for percussion
        yield from midiseq("76 77 77 77 | 77 77 77 77", channel="Standard")

    @player
    def percussion():
        # note there is no '|' on the second line, this will expand to a 8-element-line
        yield from midiseq(
            "| 44  42  46  42"
            "  44  42  46  42", channel="Standard", velocity=50)

    BD = 36 # 38 base drum
    SD = 24 # snare
    ST = 31 # 37 stick
    CH = 42 # 44 closed hat
    OH = 46 # open hat

    @player
    def drums():
        yield from midiseq(
            f"| ({BD},{CH}  .  {OH} .  {SD},{CH} . {OH}  . ) "
            f"  ({BD},{CH}  .  {CH} .  {SD},{OH} . {CH}  . ) "
            , channel="Standard", velocity=60)

    @player
    def xylophone1():
        # Paranthesis can be used to sub-divide one element in a pattern
        yield from midiseq(
            "| I>          (I> . V . ) (. . VII- .)  V "
            "| (I> . . I>) (I> . V . ) (. . VII- .)  V ",
            key=key,
            channel="Xylophone",
            velocity=40,
        )

    @player
    def xylophone2():
        yield from midiseq(
            # &se random pattern if you know how
            " ".join(
                random.choice([
                    "I>", "II>", "III>", "V>", "VI>", ".", "."
                    ]) for i in range(16)
            ),
            key=key,
            channel="Xylophone",
            velocity=40,
        )



#!/usr/bin/env python3

"""API for fancy decorator based track definitions"""

# pylint: disable=too-many-locals

import logging
import re
from collections.abc import Callable, MutableSequence, Sequence
from contextlib import contextmanager, suppress
from itertools import count
from typing import Union

from pocketrockit.engine import EMPTY_COMMAND, Env, MidiCmd, Player, Stage, run
from pocketrockit.misc import error, throw

NestedStringSequence = MutableSequence[Union[str, "NestedStringSequence"]]

stage = Stage()

__all__ = ["track", "player", "midiseq"]


def logger() -> logging.Logger:
    """Named logger"""
    return logging.getLogger("pr.decorated")


def run_engine(initial_track: str) -> bool:
    """If engine is not yet running, run it and never return"""
    if stage.active:
        return False

    stage.active = True
    stage.file_to_track = initial_track
    run()
    return True


def track(track_definition_fn: Callable[[Env], None]) -> None:
    """Creates a track (a list of note generators called players) by running provided
    @track_definition_fn
    Only effective starting from second run."""
    if run_engine(track_definition_fn.__globals__["__file__"]):
        return

    logger().info("define track '%s'", track_definition_fn.__name__)
    stage.new_track = {}
    env = Env(bpm=stage.env.bpm)
    try:
        # create the new track model
        track_definition_fn(env)
    except Exception as exc:  # pylint: disable=broad-except
        if not stage.verified:
            raise
        error(f"exception in build: {exc}")
        # In case of errors we abort loading the track, keeping the currently loaded one
        return

    stage.players, stage.new_track, stage.env = stage.new_track, None, env

    # initialize note generators (and also trigger eventual bugs!)
    for name, stream in stage.players.items():
        logger().info("initialize player %s with tick=%d", name, stage.tick)
        try:
            if not len(next(stream)) == 3:
                raise RuntimeError(f"Player '{name}' didn't return valid command")
            stream.send(stage.tick + 1)
        except Exception as exc:  # pylint: disable=broad-except
            if not stage.verified:
                raise
            # In case of errors we only skip the affected player, accepting all others
            error(f"exception in player init: {exc}")

    stage.verified = True


def player(player_definition_fn: Callable[[], Player]) -> None:
    """Adds generator created from @player_definition_fn to new track definition"""
    name = player_definition_fn.__name__
    logger().info("define player %r", name)
    assert stage.new_track is not None
    if stage.modifiers:
        stage.new_track[name] = stage.modifiers[0](player_definition_fn())
    else:
        stage.new_track[name] = player_definition_fn()


@contextmanager
def echo(delay, factor):
    """Echoes"""

    def repeat(source):
        buffer = []
        for tick, (pos, origin, cmd) in enumerate(source):
            if cmd:
                buffer.insert(0, (tick + delay, cmd))
            if buffer and buffer[-1][0] == tick:
                new_cmd = cmd + [(c, i, int(v * factor)) for c, i, v in buffer[-1][1]]
                buffer.pop()
            else:
                new_cmd = cmd

            if cmd:
                print(tick, pos, len(buffer), new_cmd)
            yield pos, origin, new_cmd
        print("ende")

    try:
        Stage().modifiers.append(repeat)
        yield
    finally:
        Stage().modifiers.pop()


def midi_scale(key: str) -> Sequence[int]:
    """Create a major or minor scale based on @key. @key can be any element of NOTES. Alphabetic
    notes ('C', 'C#', 'Db', etc.) without an octave identifier default to octave 4.
    E.g. 'C' => 'C4', etc.
    Numeric values will be directly interpreted as MIDI values
    """
    base = key.strip("-")
    minor = key[-1] == "-"
    octave = "" if any(base[i] in {"0123456789"} for i in (0, -1)) else "4"
    base_tone = NOTES.get(f"{base}{octave}", NOTES[base])

    return [
        base_tone + e for e in ([0, 2, 3, 5, 7, 8, 10, 12] if minor else [0, 2, 4, 5, 7, 9, 11, 12])
    ]


def base_note(note_str: str) -> tuple[None | str, ...]:
    """Dissects a note written in one of all possible ways into its (optional) components"""
    if match := re.match(
        r"^("
        "(((n|t)([1-8]))|([iv]{1,4}))"
        "|((a|b|c|d|e|f|g|a|h)(#|b|is|es|s)?([0-9]|[’|']*)?)"
        "|([0-9]{1,3}))"
        "([><+-]*)$",
        note_str,
        re.IGNORECASE,
    ):
        # return tuple(match.groups())
        return tuple(match.group(5, 6, 7, 11, 12))
    raise RuntimeError(note_str)


def resolve(note_str: str | int, scale: Sequence[int]) -> int:
    """Not in the mood to write a comment"""
    if isinstance(note_str, int):
        return note_str
    with suppress(ValueError):
        return int(note_str)
    degree, rom_degree, note, num_note, modifier = base_note(note_str)
    shift = (
        sum(
            12 if e == ">" else -12 if e == "<" else 1 if e == "+" else -1 if e == "-" else 0
            for e in modifier
        )
        if modifier
        else 0
    )
    tone = (
        scale[
            int(degree) - 1
            if degree is not None
            else ["i", "ii", "iii", "iv", "v", "vi", "vii", "viii"].index(rom_degree.lower())
            if rom_degree is not None
            else throw(RuntimeError())
        ]
        if degree or rom_degree
        else NOTES[note or num_note or throw(RuntimeError())]
    )

    return tone + shift


def midi_commands(
    string: str, channel: int, force_note: None | str | int, scale: Sequence[int], velocity: int
) -> Sequence[MidiCmd]:
    """Turns something like '60', 'c', 'c,d,e' or 't1,t2,t3' into MIDI commands"""
    try:
        return (
            []
            if string == "."
            else [
                (
                    channel,
                    resolve(
                        force_note if force_note is not None and element == "x" else element, scale
                    ),
                    velocity,
                )
                for element in string.split(",")
            ]
        )
    except Exception as exc:  # pylint: disable=broad-except
        if not Stage().verified:
            raise
        error(f"could not turn '{string}' into note ({exc})")

    return []


def parenthetic_contents(string: str) -> NestedStringSequence:
    """Generate parenthesized contents in string as pairs (level, contents).
    >>> list(parenthetic_contents(" a  b"))
    ['a', 'b']
    >>> list(parenthetic_contents(" ab (cd)ef( gh )if  (kl(mn))op"))
    ['ab', ['cd'], 'ef', ['gh'], 'if', ['kl', ['mn']], 'op']
    """
    back_pointer, has_content = 0, False
    current: NestedStringSequence
    current = []
    stack: MutableSequence[NestedStringSequence] = [current]

    for i, char in enumerate(string):
        if char == " ":
            if has_content:
                has_content = False
                current.append(string[back_pointer:i])
        elif char == "(":
            if has_content:
                current.append(string[back_pointer:i])
            has_content = False
            stack.append(current)
            new_element: NestedStringSequence = []
            current.append(current := new_element)
        elif char == ")":
            if has_content:
                current.append(string[back_pointer:i])
            has_content = False
            current = stack.pop()
        else:
            if not has_content:
                back_pointer, has_content = i, True
    if has_content:
        current.append(string[back_pointer:])

    # print(string, stack[0])
    return stack[0]


def quantized_pattern(
    nested_pattern: NestedStringSequence, resolution: int
) -> Sequence[tuple[int, str]]:
    """Assigns a line-relative tick number to every note"""
    result: MutableSequence[tuple[int, str]] = []
    # assert len(nested_pattern)
    # assert depth(nested_pattern)
    # assert sizes are modules

    step_size = resolution // len(nested_pattern)
    for i, elem in enumerate(nested_pattern):
        if isinstance(elem, str):
            result.append((i * step_size, elem))
        else:
            sub_step_size = step_size // len(elem)
            for j, sub_elem in enumerate(elem):
                result.append((i * step_size + j * sub_step_size, str(sub_elem)))
    # print(result)
    return result


def pattern_list_from(
    *pattern_definition: str | Sequence[str],
) -> Sequence[Sequence[Sequence[tuple[int, str]]]]:
    """Turns various pattern definition styles into"""

    def block(pattern: str) -> Sequence[Sequence[tuple[int, str]]]:
        return [
            quantized_pattern(parenthetic_contents(line), Stage().env.step_size)
            for line in pattern.split("|")
            if line.strip()
        ]

    if all(isinstance(p, str) for p in pattern_definition):
        # only one string or a tuple of strings
        return [block(str(p)) for p in pattern_definition]

    if isinstance(pattern_definition[0], (list, tuple)) and len(pattern_definition) == 1:
        return [block(p) for p in pattern_definition[0]]

    raise RuntimeError(f"Did not understand pattern definition {pattern_definition}")


def midiseq(
    *pattern: str | Sequence[str],
    channel: int = 0,
    note: None | int | str = None,
    key: str | Sequence[str] = "C4",
    velocity: int = 30,
    **_kwargs: str | int,
) -> Player:
    """A minimal MIDI sequencer"""
    pattern_blocks = pattern_list_from(*pattern)
    step_size = Stage().env.step_size
    loop_size = step_size * sum(len(elements) for elements in pattern_blocks)

    reset_tick: None | int = None
    keys = key.split(",") if isinstance(key, str) else key

    while True:
        if reset_tick is not None:
            logger().debug("reset to %d", reset_tick)
            yield EMPTY_COMMAND

        start_loop = (reset_tick or 0) // loop_size
        skip = (reset_tick or 0) % loop_size
        tick_iter = count((reset_tick or 0) - skip)
        reset_tick = None

        for loop_nr, block_nr, _block_key, scale, line_nr, (next_relative_tick, next_element) in (
            (_loop_nr, _block_nr, _key, _scale, _line_nr, _element)
            for _loop_nr in count(start_loop)
            for _line_nr, (_block_nr, _key, _scale, _line) in enumerate(
                (__block_nr, __key, __scale, __line)
                for __block_nr, __lines in enumerate(pattern_blocks)
                for __key in (keys[(_loop_nr * len(pattern_blocks) + __block_nr) % len(keys)],)
                for __scale in (midi_scale(__key),)
                for __line in __lines
            )
            for _element in _line
        ):

            next_tick = loop_nr * loop_size + line_nr * step_size + next_relative_tick
            # logger().debug(
            #     "loop: %d block: %d key: %s line: %d next_tick: %d => %s",
            #     loop_nr,
            #     block_nr,
            #     block_key,
            #     line_nr,
            #     next_tick,
            #     next_element,
            # )
            for tick in tick_iter:
                play = tick == next_tick

                if skip:
                    skip -= 1
                else:
                    assert next_tick >= tick
                    reset_tick = yield (
                        (
                            (loop_nr, block_nr, tick),
                            next_element,
                            midi_commands(next_element, channel, note, scale, velocity),
                        )
                        if play
                        else EMPTY_COMMAND
                    )

                if play:
                    break

            if reset_tick is not None:
                break


# B2!!
# A2!!
NOTES = {
    name: midi
    for midi, names, freq in (
        (21, {"21", "A0"}, 27.5),  # A2!!
        (22, {"Bb0", "22", "Ais2", "A#0"}, 29.14),
        (23, {"B0", "23", "H2"}, 30.87),
        (24, {"24", "B#0", "C1"}, 32.7),
        (25, {"C#1", "Cis1", "Des1", "Db1", "25"}, 34.65),
        (26, {"26", "D1"}, 36.71),
        (27, {"Es1", "27", "Dis1", "D#1", "Eb1"}, 38.89),
        (28, {"E1", "28"}, 41.2),
        (29, {"F1", "E#1", "29"}, 43.65),
        (30, {"F#1", "Gb1", "Fis1", "Ges1", "30"}, 46.25),
        (31, {"G1", "31"}, 49.0),
        (32, {"Gis1", "As1", "Ab1", "G#1", "32"}, 51.91),
        (33, {"33", "A1"}, 55.0),
        (34, {"b1", "A#1", "Bb1", "34", "Ais1"}, 58.27),
        (35, {"B1", "35", "H1"}, 61.74),
        (36, {"C", "C2", "B#1", "36"}, 65.41),
        (37, {"37", "C#2", "Db2", "Cis", "Des"}, 69.3),
        (38, {"D", "38", "D2"}, 73.42),
        (39, {"Eb2", "D#2", "Dis", "Es", "39"}, 77.78),
        (40, {"E2", "E", "40"}, 82.41),
        (41, {"F", "E#2", "41", "F2"}, 87.31),
        (42, {"F#2", "Ges", "42", "Gb2", "Fis"}, 92.5),
        (43, {"G2", "G", "43"}, 98.0),
        (44, {"44", "G#2", "Ab2", "As", "Gis"}, 103.83),
        (45, {"A", "45", "A2"}, 110.0),
        (46, {"Ais", "A#2", "Bb2", "B", "46"}, 116.54),
        (47, {"47", "H", "B2"}, 123.47),
        (48, {"C3", "B#2", "c", "48"}, 130.81),
        (49, {"Db3", "cis", "C#3", "49", "des"}, 138.59),
        (50, {"D3", "d", "50"}, 146.83),
        (51, {"es", "dis", "D#3", "51", "Eb3"}, 155.56),
        (52, {"E3", "e", "52"}, 164.81),
        (53, {"f", "E#3", "53", "F3"}, 174.61),
        (54, {"ges", "Gb3", "54", "fis", "F#3"}, 185.0),
        (55, {"55", "g", "G3"}, 196.0),
        (56, {"56", "G#3", "as", "gis", "Ab3"}, 207.65),
        (57, {"57", "A3", "a"}, 220.0),
        (58, {"Bb3", "b", "58", "A#3", "ais"}, 233.08),
        (59, {"h", "B3", "59"}, 246.94),
        (60, {"c’", "60", "B#3", "C4"}, 261.63),  # Middle C
        (61, {"Db4", "C#4", "des’", "cis’", "61"}, 277.18),
        (62, {"D4", "d’", "62"}, 293.66),
        (63, {"es’", "D#4", "63", "Eb4", "dis’"}, 311.13),
        (64, {"e’", "64", "E4"}, 329.63),
        (65, {"f’", "65", "E#4", "F4"}, 349.23),
        (66, {"fis’", "ges’", "Gb4", "F#4", "66"}, 369.99),
        (67, {"g’", "67", "G4"}, 392.0),
        (68, {"Ab4", "as’", "68", "gis’", "G#4"}, 415.3),
        (69, {"69", "A4", "a’"}, 440.0),
        (70, {"70", "A#4", "Bb4", "ais’", "b’"}, 466.16),
        (71, {"h’", "71", "B4"}, 493.88),
        (72, {"72", "c’’", "B#4", "C5"}, 523.25),
        (73, {"C#5", "cis’’", "73", "Db5", "des’’"}, 554.37),
        (74, {"D5", "74", "d’’"}, 587.33),
        (75, {"dis’’", "Eb5", "es’’", "D#5", "75"}, 622.25),
        (76, {"e’’", "76", "E5"}, 659.26),
        (77, {"f’’", "77", "E#5", "F5"}, 698.46),
        (78, {"fis’’", "ges’’", "F#5", "78", "Gb5"}, 739.99),
        (79, {"79", "G5", "g’’"}, 783.99),
        (80, {"as’’", "gis’’", "80", "Ab5", "G#5"}, 830.61),
        (81, {"81", "A5", "a’’"}, 880.0),
        (82, {"A#5", "82", "b’’", "ais’’", "Bb5"}, 932.33),
        (83, {"h’’", "83", "B5"}, 987.77),
        (84, {"c’’’", "84", "B#5", "C6"}, 1046.5),
        (85, {"cis’’’", "des’’’", "C#6", "Db6", "85"}, 1108.73),
        (86, {"D6", "86", "d’’’"}, 1174.66),
        (87, {"D#6", "dis’’’", "87", "Eb6", "es’’’"}, 1244.51),
        (88, {"E6", "88", "e’’’"}, 1318.51),
        (89, {"f’’’", "F6", "E#6", "89"}, 1396.91),
        (90, {"fis’’’", "F#6", "ges’’’", "Gb6", "90"}, 1479.98),
        (91, {"91", "g’’’", "G6"}, 1567.98),
        (92, {"as’’’", "gis’’’", "G#6", "Ab6", "92"}, 1661.22),
        (93, {"A6", "a’’’", "93"}, 1760.0),
        (94, {"A#6", "94", "Bb6", "b’’’", "ais’’’"}, 1864.66),
        (95, {"95", "h’’’", "B6"}, 1975.53),
        (96, {"c’’’’", "C7", "B#6", "96"}, 2093.0),
        (97, {"97", "Db7", "C#7", "des’’’’", "cis’’’’"}, 2217.46),
        (98, {"d’’’’", "98", "D7"}, 2349.32),
        (99, {"D#7", "dis’’’’", "Eb7", "99", "es’’’’"}, 2489.02),
        (100, {"E7", "100", "e’’’’"}, 2637.02),
        (101, {"f’’’’", "E#7", "F7", "101"}, 2793.83),
        (102, {"fis’’’’", "ges’’’’", "Gb7", "F#7", "102"}, 2959.96),
        (103, {"103", "g’’’’", "G7"}, 3135.96),
        (104, {"Ab7", "G#7", "as’’’’", "104", "gis’’’’"}, 3322.44),
        (105, {"105", "a’’’’", "A7"}, 3520.0),
        (106, {"106", "Bb7", "ais’’’’", "A#7", "b’’’’"}, 3729.31),
        (107, {"h’’’’", "107", "B7"}, 3951.07),
        (108, {"c’’’’’", "108", "B#7", "C8"}, 4186.01),
        (109, {"109", "Db8", "des’’’’’", "cis’’’’’", "C#8"}, 4434.92),
        (110, {"d’’’’’", "D8", "110"}, 4698.64),
        (111, {"dis’’’’’", "Eb8", "D#8", "111", "es’’’’’"}, 4978.03),
        (112, {"112", "e’’’’’", "E8"}, 5274.04),
        (113, {"F8", "E#8", "113", "f’’’’’"}, 5587.65),
        (114, {"Gb8", "F#8", "ges’’’’’", "fis’’’’’", "114"}, 5919.91),
        (115, {"g’’’’’", "G8", "115"}, 6271.93),
        (116, {"116", "Ab8", "as’’’’’", "G#8", "gis’’’’’"}, 6644.88),
        (117, {"117", "a’’’’’", "A8"}, 7040.0),
        (118, {"Bb8", "b’’’’’", "A#8", "118", "ais’’’’’"}, 7458.62),
        (119, {"119", "h’’’’’", "B8"}, 7902.13),
        (120, {"c’’’’’’", "C9", "B#8", "120"}, 8372.02),
        (121, {"Db9", "121", "C#9", "des’’’’’’", "cis’’’’’’"}, 8869.84),
        (122, {"d’’’’’’", "122", "D9"}, 9397.27),
        (123, {"dis’’’’’’", "es’’’’’’", "123", "D#9", "Eb9"}, 9956.06),
        (124, {"E9", "e’’’’’’", "124"}, 10548.08),
        (125, {"F9", "E#9", "f’’’’’’", "125"}, 11175.3),
        (126, {"fis’’’’’’", "126", "ges’’’’’’", "Gb9", "F#9"}, 11839.82),
        (127, {"G9", "g’’’’’’", "127"}, 12543.85),
    )
    for name in names
}



#!/usr/bin/env python3

"""Stuff that doesn't go anywhere else
"""

import asyncio
import logging
import os
import sys
import termios
import threading
from collections.abc import AsyncIterator, Callable, Coroutine, Iterator, MutableMapping
from contextlib import contextmanager
from functools import wraps
from typing import NoReturn, TextIO


def logger() -> logging.Logger:
    """Logger for this module"""
    return logging.getLogger("pr.misc")


class Singleton(type):
    """Yes, a Singleton"""

    _instances: MutableMapping[type, object] = {}

    def __call__(cls: "Singleton", *args: object, **kwargs: object) -> object:
        """Creates an instance if not available yet, returns it"""
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


def watchdog(
    afunc: Callable[..., Coroutine[object, object, object]]
) -> Callable[..., Coroutine[object, object, object]]:
    """Watch for async functions to throw an unhandled exception"""

    @wraps(afunc)
    async def run(*args: object, **kwargs: object) -> object:
        """Run wrapped function and handle exceptions"""
        try:
            return await afunc(*args, **kwargs)
        except asyncio.CancelledError:
            logger().info("Task cancelled: `%s`", afunc.__name__)
        except KeyboardInterrupt:
            logger().info("KeyboardInterrupt in `%s`", afunc.__name__)
        except Exception:  # pylint: disable=broad-except
            logger().exception("Exception in `%s`:", afunc.__name__)
            asyncio.get_event_loop().stop()
        return None

    return run


def colored(text: str, color: str) -> str:
    """Returns @color formatted @text"""
    return (
        text
        if color is None or not sys.stdout.isatty() or "COLORTERM" not in os.environ
        else {
            "none": "%s",
            "green": "\033[0;32m%s\033[0m",
            "green_bold": "\033[1;32m%s\033[0m",
            "red": "\033[0;31m%s\033[0m",
            "red_bold": "\033[1;31m%s\033[0m",
            "black": "\033[0;30m%s\033[0m",
            "black_bold": "\033[1;30m%s\033[0m",
            "blue": "\033[0;34m%s\033[0m",
            "blue_bold": "\033[1;34m%s\033[0m",
            "yellow": "\033[0;33m%s\033[0m",
            "yellow_bold": "\033[1;33m%s\033[0m",
        }.get(color, "%s")
        % text
    )


def error(message: str) -> None:
    """Prints colorful text"""
    print(colored(message, "red"))


@contextmanager
def raw_mode(file: TextIO) -> Iterator[None]:
    """Context manager switches off stream buffering and resets it aftwerwards"""
    old_attrs = termios.tcgetattr(file.fileno())
    new_attrs = old_attrs[:]
    new_attrs[3] = new_attrs[3] & ~(termios.ECHO | termios.ICANON)
    try:
        termios.tcsetattr(file.fileno(), termios.TCSADRAIN, new_attrs)
        yield
    finally:
        termios.tcsetattr(file.fileno(), termios.TCSADRAIN, old_attrs)


async def keyboard_reader(
    loop: asyncio.BaseEventLoop, _terminator: threading.Event
) -> AsyncIterator[str]:
    """Async keyboard reader"""
    try:
        with raw_mode(sys.stdin):
            reader = asyncio.StreamReader()
            await loop.connect_read_pipe(lambda: asyncio.StreamReaderProtocol(reader), sys.stdin)
            while not reader.at_eof():
                # handle terminator         while not terminator.is_set():
                if (char := bytes.decode(await reader.read(2)).replace("\r", "\n")) == chr(4):
                    # chr(4) means EOT (sent by CTRL+D on UNIX terminals)
                    logger().info("CTRL+D")
                    break
                yield char
    except termios.error as exc:
        raise RuntimeError(f"Could not use raw mode on STDIN: {exc}") from exc
    finally:
        logging.debug("keyboard reader terminated")


def setup_logging(level: str | int = logging.DEBUG) -> None:
    '''
    def thread_id_filter(record):
        """Inject thread_id to log records"""
        record.thread_id = threading.get_native_id()
        return record

    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter("%(asctime)s | %(levelname)s | %(thread_id)s | %(message)s")
    )
    handler.addFilter(thread_id_filter)
    logger().addHandler(handler)
    logging.getLogger().setLevel(level)
    '''
    use_col = "TERM" in os.environ
    col_terminator = "\033[0m" if use_col else ""
    logging.basicConfig(
        format=f"%(levelname)s %(asctime)s.%(msecs)03d %(name)-12s│ %(message)s{col_terminator}",
        datefmt="%H:%M:%S",
        level=getattr(logging, level) if isinstance(level, str) else level,
    )
    for name, color in (
        ("DEBUG", "\033[32m"),
        ("INFO", "\033[36m"),
        ("WARNING", "\033[33m"),
        ("ERROR", "\033[31m"),
        ("CRITICAL", "\033[37m"),
    ):
        logging.addLevelName(
            getattr(logging, name),
            f"{color if use_col else ''}({name[0] * 2})",
        )


def throw(exc: Exception) -> NoReturn:
    """Make raising an exception functional"""
    raise exc
